/*=========================================================================
 * perm.c
 * Computes the permanent of a (complex) square NxN matrix.
 *
 * Inspired by and improves on the implementation by Brian K. Butler. It 
 * is slightly faster and does support complex matrices.
 *
 * This algorithm implements a variant of Ryser's algorithm with a factor 2
 * speed-up, as explained by A. Nijenhuis and H. S. Wilf in their second 
 * edition of "Combinatorial Algorithms" (1978). See:
 * https://www.math.upenn.edu/~wilf/website/CombAlgDownld.html
 * 
 * This is a MEX-file for MATLAB.
 * Matthijs Velsink, 2020
 *=======================================================================*/

#include "mex.h"
#include <stdint.h>
#include <stdlib.h>

// Define mxComplexDouble in case of < R2018a
#if MX_HAS_INTERLEAVED_COMPLEX
#else
    typedef struct {mxDouble real, imag;} mxComplexDouble;
#endif
// Define input/output and indexing macro
#define INPUT (prhs[0])
#define OUTPUT (plhs[0])
#define A(i, j) (A[(i) + (j) * N])
                                         
/* Complex arithmetic. */  
static inline mxComplexDouble CompAdd(mxComplexDouble a, mxComplexDouble b) {
    return (mxComplexDouble) {a.real + b.real, a.imag + b.imag};
}
static inline mxComplexDouble CompSub(mxComplexDouble a, mxComplexDouble b) {
    return (mxComplexDouble) {a.real - b.real, a.imag - b.imag};
}
static inline mxComplexDouble CompMul(mxComplexDouble a, mxComplexDouble b) {
    return (mxComplexDouble) {a.real * b.real - a.imag * b.imag, 
                              a.real * b.imag + a.imag * b.real};
}                                         
static inline mxComplexDouble CompSca(double a, mxComplexDouble b) {
    return (mxComplexDouble) {a * b.real, a * b.imag};
}      

/* Permanent for real matrix. */
mxDouble RealPerm(mxDouble const *const A, int N) {
    
    uint64_t maxS = (1ULL << (N - 1)) - 1; // Final permutation (all ones)
    uint64_t prevGray = 0;              // Previous Gray code
    double x[64];                       // Row sum (see Nij. & Wi.)    
    double sum = 1.0;                   // Running sum over all S
    
    // Calculate added column values and add product for the zero set
    for (int i = 0; i < N; i++) {
        double Ai_sum_j = A(i, 0);
        for (int j = 1; j < N; j++) {
            Ai_sum_j += A(i, j);
        }
        x[i] = A(i, N - 1) - Ai_sum_j/2.0;
        sum *= x[i];
    }
    
    // Go over all other permutations of S (we already did the S = {})
    for (uint64_t S = 1; S <= maxS; S++) {
        // Calculate new Gray code and difference with previous with xor
        uint64_t Gray = S^(S >> 1);
        uint64_t GrayFlip = Gray^prevGray;
        prevGray = Gray;
        // Find corresponding flipped index by counting trailing zeros
        int j = __builtin_ctzll(GrayFlip);
        
        // Add or subtract flipped row and multiply
        double product = 1.0;
        double *xptr = x;               // Pointers for reduced indexing
        double const *Aptr = &A(0, j);  // Pointers for reduced indexing
        if (GrayFlip&Gray) {
            // Row bit turned on, so add
            for (int i = 0; i < N; i++) {
                // x[i] += A(i, j);
                *xptr += *Aptr++;   
                // product *= x[i];
                product *= *xptr++; 
            }            
        } else {
            // Row bit turned off, so subtract
            for (int i = 0; i < N; i++) {
                // x[i] += A(i, j);
                *xptr -= *Aptr++;   
                // product *= x[i];
                product *= *xptr++; 
            }
        }
        // Add or subtract product, based on even/odd of S
        if (S&1) {
            sum -= product;
        } else {
            sum += product;
        }
    }
    
    // Prefactor multiplication, based on even/odd of N - 1
    if (N&1) {
        return 2.0 * sum;
    } else {
        return -2.0 * sum;
    }
}

/* Permanent for complex matrix. */
mxComplexDouble ComplexPerm(mxComplexDouble const *const A, int N) {
    
    uint64_t maxS = (1ULL << (N - 1)) - 1; // Final permutation (all ones)
    uint64_t prevGray = 0;              // Previous Gray code
    mxComplexDouble x[28];              // Row sum (see Nij. & Wi.)    
    mxComplexDouble sum = {1.0, 0.0};   // Running sum over all S    

    // Calculate added column values and add product for the zero set
    for (int i = 0; i < N; i++) {
        mxComplexDouble Ai_sum_j = A(i, 0);
        for (int j = 1; j < N; j++) {
        	Ai_sum_j = CompAdd(Ai_sum_j, A(i, j));
        }
        x[i] = CompSub(A(i, N - 1), CompSca(0.5, Ai_sum_j));
        sum = CompMul(sum, x[i]);
    }
    
    // Go over all other permutations of S (we already did the S = {})
    for (uint64_t S = 1; S <= maxS; S++) {
        // Calculate new Gray code and difference with previous with xor
        uint64_t Gray = S^(S >> 1);
        uint64_t GrayFlip = Gray^prevGray;
        prevGray = Gray;
        // Find corresponding flipped index by counting trailing zeros
        int j = __builtin_ctzll(GrayFlip);
        
        // Add or subtract flipped row and multiply
        mxComplexDouble product = {1.0, 0.0};
        mxComplexDouble *xptr = x;              // Pointers for reduced indexing
        mxComplexDouble const *Aptr = &A(0, j); // Pointers for reduced indexing
        if (GrayFlip&Gray) {
            // Row bit turned on, so add
            for (int i = 0; i < N; i++) {
                // x[i] = CompAdd(x[i], A(i, j));
                *xptr = CompAdd(*xptr, *Aptr);
                Aptr++;
                // product = CompMul(product, x[i]);
                product = CompMul(product, *xptr);
                xptr++;
            }            
        } else {
            // Row bit turned off, so subtract
            for (int i = 0; i < N; i++) {
                // x[i] = CompSub(x[i], A(i, j));
                *xptr = CompSub(*xptr, *Aptr);
                Aptr++;
                // product = CompMul(product, x[i]);
                product = CompMul(product, *xptr);
                xptr++;
            }
        }
        // Add or subtract product, based on even/odd of S
        if (S&1) {
            sum = CompSub(sum, product);
        } else {
            sum = CompAdd(sum, product);
        }
    }
    
    // Prefactor multiplication, based on even/odd of N - 1
    if (N&1) {
        return CompSca(2.0, sum);
    } else {
        return CompSca(-2.0, sum);
    }
}

/* The gateway routine. */
void mexFunction(int nlhs, mxArray *plhs[], 
                 int nrhs, mxArray const *prhs[]) {
    
    // Validate number of inputs/outputs
    if (nrhs != 1) {
        mexErrMsgIdAndTxt("perm:rhs",
            "This function requires 1 input argument.");
    } else if (nlhs > 1) {
        mexErrMsgIdAndTxt("perm:lhs",
            "This function only gives 1 output argument.");
    }
    // Validate matrix type
    if (!mxIsDouble(INPUT) || mxIsSparse(INPUT)) {
        mexErrMsgIdAndTxt("perm:inputMatrixType",
            "Input argument must be a full matrix of (complex) doubles.");
    }
    // Validate matrix size
    int DimNum = (int)mxGetNumberOfDimensions(INPUT);
    const mwSize *DimSize = mxGetDimensions(INPUT); 
    int N = (int)DimSize[0];
    if ((DimSize[0] != DimSize[1]) || (DimNum > 2) || (N > 64)){                              
        mexErrMsgIdAndTxt("perm:inputMatrixSize",
            "Input argument must be a square matrix not bigger than 64x64.");
    }
    
    #if MX_HAS_INTERLEAVED_COMPLEX
    // Calculate permanent with direct interleaved complex API (>= R2018a)
    if (mxIsComplex(INPUT)) {
        // Initialize complex output and run complex perm function
        OUTPUT = mxCreateDoubleMatrix(1, 1, mxCOMPLEX);        
        if (N == 0) {
            mxGetComplexDoubles(OUTPUT)[0] = (mxComplexDouble) {1.0, 0.0};  // 0x0 matrix has permanent of 1
        } else {
            mxGetComplexDoubles(OUTPUT)[0] = ComplexPerm(mxGetComplexDoubles(INPUT), N);
        }
    } else {
        // Initialize real output and run real perm function
        OUTPUT = mxCreateDoubleMatrix(1, 1, mxREAL);
        if (N == 0) {
            mxGetDoubles(OUTPUT)[0] = 1.0;  // 0x0 matrix has permanent of 1
        } else {            
            mxGetDoubles(OUTPUT)[0] = RealPerm(mxGetDoubles(INPUT), N);
        }
    }
    #else
    // Calculate permanent by first converting to interleaved complex (< R2018a)    
    if (mxIsComplex(INPUT)) {
        // Initialize complex output and run complex perm function
        OUTPUT = mxCreateDoubleMatrix(1, 1, mxCOMPLEX);
        if (N == 0) {
            mxGetPr(OUTPUT)[0] = 1.0;  // 0x0 matrix has permanent of 1
            mxGetPi(OUTPUT)[0] = 0.0;
        } else {
            // Convert matrix to interleaved (negligible overhead)
            mxDouble *APr = mxGetPr(INPUT);
            mxDouble *APi = mxGetPi(INPUT);
            mxComplexDouble *A = malloc(sizeof(*A) * (size_t)N * (size_t)N);
            for (int i = 0; i < N*N; i++) {
                A[i].real = APr[i];
                A[i].imag = APi[i];
            }
            mxComplexDouble result = ComplexPerm(A, N);
            free(A);
            mxGetPr(OUTPUT)[0] = result.real;
            mxGetPi(OUTPUT)[0] = result.imag;            
        }
    } else {
        // Initialize real output and run real perm function
        OUTPUT = mxCreateDoubleMatrix(1, 1, mxREAL);
        if (N == 0) {            
            mxGetPr(OUTPUT)[0] = 1.0;  // 0x0 matrix has permanent of 1
        } else {
            mxGetPr(OUTPUT)[0] = RealPerm(mxGetPr(INPUT), N);
        }
    }
    #endif
}