
function [pdf, labels] = getBSpdf(inputs, U, x, mode_cls, mode_dsc)

% This function computes the probability distrifution function of a Boson
% Sampler according to Renema, Phys. Rev. Lett. 120, 220502, (2018)
% Coded by Malaquias Correa Anguita (m.correaanguita@utwente.nl)
% NB: Not efficient.
% 
% Arguments:
% - inputs              : 1-d array containing the input modes of n photons to be injected into the interferometer
% - U                   : 2-d unitary matrix that describes the transformation between input/output modes of the interferometer
% - x (optional)        : Partial distinguishability between photons (assumed to be equal for all photon pairs)
%                         default x=1
% - mode_cls (optional) : Collision mode. Select 1 (True) to allow collision, 0 (False) otherwise
%                         default mode_cls = 0
% - mode_dsc (optional) : Descending order mode. Select 1 (True) to sort detection events of
%                         pdf in desceding order
%                         default mode_dsc = 1
% 
% Outputs;
% - pdf    : 1-d array containing the normalized probabilities for each possible detection event
% - labels : array of strings containing the labels of the detection events

% Defining default values
if nargin < 2
    error('Not enough input arguments.')
elseif nargin == 2
    x = 1;
    mode_cls = 0;
    mode_dsc = 1;
elseif nargin == 3
    mode_cls = 0;
    mode_dsc = 1;
elseif nargin == 4
    mode_dsc = 1;
end
    
n = numel(inputs); % n photons
m = size(U,1);     % m mode interferometer

if size(U,1) ~= size(U,2)
    error("Please enter square matrix U")
end

% Partial distinguishability powers for expansion
xpowers = x.^(0:n);

% 
combs = bsxfun(@minus, nchoosek(1:n+m-1,n), 0:n-1);

if mode_cls == 0
    % This part of the code removes microstates which have collisions
    indexes_delete = [];
    for jj = 1:length(combs(:,1))
        if length(unique(combs(jj,:))) ~= length(combs(1,:))
            indexes_delete(end + 1) = jj;
        end
    end
    combs(indexes_delete, :) = [];
end
length(combs(:,1));

%auxiliary array to store results
res = [];

for ii = 1:size(combs,1)
    label = '';
    for ch = 1:n
        label = append(label, num2str(combs(ii,ch)));
    end
    label = convertCharsToStrings(label);
    
    coeffs = BSprobwdist(U(inputs,combs(ii,:)));
    
    % Exact Prob Dist
    res(ii,1) = abs(sum(coeffs(end:-1:1).*xpowers)) ...
        * size(unique(perms(combs(ii,:)),'rows'),1) / factorial(n);
    
    res(ii,2) = label;

    
end
if mode_dsc == 1
    res = sortrows(res, 'descend');
end

%Prepare the outputs
pdf = res(:,1)/sum(res(:,1));
labels = res(:,2);
    
 