/*=========================================================================
 * permT.c
 * Computes the permanent of a (complex) NxNxN tensor.
 *
 * Similar to my implementation of the permanent of a (complex) NxN matrix.
 *
 * This algorithm implements a variant of Ryser's algorithm with a factor 2
 * speed-up, as explained by A. Nijenhuis and H. S. Wilf in their second 
 * edition of "Combinatorial Algorithms" (1978). See:
 * https://www.math.upenn.edu/~wilf/website/CombAlgDownld.html
 * 
 * This is a MEX-file for MATLAB.
 * Matthijs Velsink, 2020
 *=======================================================================*/

#include "mex.h"
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

// Define mxComplexDouble in case of < R2018a
#if MX_HAS_INTERLEAVED_COMPLEX
#else
    typedef struct {mxDouble real, imag;} mxComplexDouble;
#endif
// Define input/output and indexing macros
#define INPUT (prhs[0])
#define OUTPUT (plhs[0])
#define W(i, j, k) (W[(i) + (j) * N + (k) * N * N])
#define y(i, k) (y[(i) + (k) * N])  
#define xMat(i, j) (xMat[(i) + (j) * N])  
    
/* Complex arithmetic. */  
static inline mxComplexDouble CompAdd(mxComplexDouble a, mxComplexDouble b) {
    return (mxComplexDouble) {a.real + b.real, a.imag + b.imag};
}
static inline mxComplexDouble CompSub(mxComplexDouble a, mxComplexDouble b) {
    return (mxComplexDouble) {a.real - b.real, a.imag - b.imag};
}
static inline mxComplexDouble CompMul(mxComplexDouble a, mxComplexDouble b) {
    return (mxComplexDouble) {a.real * b.real - a.imag * b.imag, 
                              a.real * b.imag + a.imag * b.real};
}                                         
static inline mxComplexDouble CompSca(double a, mxComplexDouble b) {
    return (mxComplexDouble) {a * b.real, a * b.imag};
}     

/* Permanent for real matrix. */
mxDouble RealPermT(mxDouble const *const W, int N) {
    
    uint64_t maxR = (1ULL << N) - 1;       // Final permutation of R (all ones) 
    uint64_t maxS = (1ULL << (N - 1)) - 1; // Final permutation of S (all ones)
    uint64_t prevGrayR = 0;                // Previous Gray code of 2nd index
    uint64_t prevGrayS = 0;                // Previous Gray code of 2nd index
    double *x = calloc((size_t)N, sizeof(*x)); // Temporary sum over j (see derivation) 
    double *xS = calloc((size_t)N, sizeof(*xS)); // x as modified by S
    double *y = calloc((size_t)N * (size_t)N, sizeof(*y)); // Temporary matrix sum over j (see derivation) 
    double *xMat = calloc((size_t)N * (size_t)N, sizeof(*xMat)); // Helper matrix for x
    double sumS = 0.0;                     // Partial sum over S
    double sum = 0.0;                      // Running sum over all R, S
    
    // Calculate weighted sum over all k (needed for x)
    for (int i = 0; i < N; i++) {
        for (int j = 0; j < N; j++) {
            double Wij_sum_k = W(i, j, 0);            
            for (int k = 1; k < N; k++) {
                Wij_sum_k += W(i, j, k);
            }  
            xMat(i, j) = W(i, j, N - 1) - 0.5*Wij_sum_k;
        }
    }
        
    // Go over all permutations of R 
    for (uint64_t R = 1; R <= maxR; R++) {
        // Calculate new Gray code and difference with previous with xor
        uint64_t GrayR = R^(R >> 1);
        uint64_t GrayFlipR = GrayR^prevGrayR;
        prevGrayR = GrayR;
        // Find corresponding flipped index by counting trailing zeros
        int j = __builtin_ctzll(GrayFlipR);
        
        // Change x and y accordingly
        double productS0 = 1.0;         // Partial sum for S = 0
        double *xptr = x;               // Pointers for reduced indexing
        double *yptr = y;               // Pointers for reduced indexing
        double *xMatptr = &xMat(0, j);  // Pointers for reduced indexing        
        if (GrayFlipR&GrayR) {
            // Bit turned on, so add to x
            for (int i = 0; i < N; i++) {
                // x[i] += xMat(i, j);
                *xptr += *xMatptr++;   
                // productS0 *= x[i];
                productS0 *= *xptr++; 
            }
            // And y
            for (int k = 0; k < N; k++) {
                for (int i = 0; i < N; i++) {
                    // y(i, k) += W(i, j, k);
                    *yptr += W(i, j, k);  
                    yptr++;                     
                }                
            }
        } else {            
            // Bit turned off, so subtract from x
            for (int i = 0; i < N; i++) {
                // x[i] -= xMat(i, j);
                *xptr -= *xMatptr++; 
                // productS0 *= x[i];
                productS0 *= *xptr++;
            }
            // And y
            for (int k = 0; k < N; k++) {
                for (int i = 0; i < N; i++) {
                    // y(i, k) -= W(i, j, k);
                    *yptr -= W(i, j, k);  
                    yptr++;
                }                
            }
        } 
        
        // Go over all permutations of S
        memcpy(xS, x, sizeof(*x)*(size_t)N);    // Allow S to modify x
        prevGrayS = 0;                      
        sumS = productS0;                       // Start with S = {} contribution  
        for (uint64_t S = 1; S <= maxS; S++) {
            // Calculate new Gray code and difference with previous with xor
            uint64_t GrayS = S^(S >> 1);
            uint64_t GrayFlipS = GrayS^prevGrayS;
            prevGrayS = GrayS;
            // Find corresponding flipped index by counting trailing zeros
            int k = __builtin_ctzll(GrayFlipS);

            // Add or subtract flipped row and multiply
            double product = 1.0;
            double *xSptr = xS;       // Pointers for reduced indexing
            double *yptr = &y(0, k);  // Pointers for reduced indexing
            if (GrayFlipS&GrayS) {
                // Bit turned on, so add
                for (int i = 0; i < N; i++) {
                    // x[i] += y(i, k);
                    *xSptr += *yptr++;   
                    // product *= x[i];
                    product *= *xSptr++; 
                }            
            } else {
                // Bit turned off, so subtract
                for (int i = 0; i < N; i++) {
                    // x[i] -= y(i, k);
                    *xSptr -= *yptr++;   
                    // product *= x[i];
                    product *= *xSptr++; 
                }   
            }
            // Add or subtract product, based on even/odd of S
            if (S&1) {
                sumS -= product;
            } else {
                sumS += product;
            }
        }
        
        // Add or subtract partial sum, based on even/odd of R
        if (R&1) {
            sum -= sumS;
        } else {
            sum += sumS;
        }        
    }
    
    free(x);
    free(xS);
    free(y);
    free(xMat);
    
    // Final -2 multiplication
    return -2.0 * sum;
}

/* Permanent for complex matrix. */
mxComplexDouble ComplexPermT(mxComplexDouble const *const W, int N) {
    
    uint64_t maxR = (1ULL << N) - 1;        // Final permutation of R (all ones) 
    uint64_t maxS = (1ULL << (N - 1)) - 1;  // Final permutation of S (all ones)
    uint64_t prevGrayR = 0;                 // Previous Gray code of 2nd index
    uint64_t prevGrayS = 0;                 // Previous Gray code of 2nd index
    mxComplexDouble *x = calloc((size_t)N, sizeof(*x)); // Temporary sum over j (see derivation) 
    mxComplexDouble *xS = calloc((size_t)N, sizeof(*xS)); // x as modified by S
    mxComplexDouble *y = calloc((size_t)N * (size_t)N, sizeof(*y)); // Temporary matrix sum over j (see derivation) 
    mxComplexDouble *xMat = calloc((size_t)N * (size_t)N, sizeof(*xMat)); // Sum over all k (times -0.5)
    mxComplexDouble sumS = {0.0, 0.0};      // Partial sum over S
    mxComplexDouble sum = {0.0, 0.0};       // Running sum over all R, S
    
    // Calculate weighted sum over all k (needed for x)
    for (int i = 0; i < N; i++) {
        for (int j = 0; j < N; j++) {
            mxComplexDouble Wij_sum_k = W(i, j, 0);            
            for (int k = 1; k < N; k++) {
                Wij_sum_k = CompAdd(Wij_sum_k, W(i, j, k));
            }  
            xMat(i, j) = CompSub(W(i, j, N - 1), CompSca(0.5, Wij_sum_k));
        }
    }
        
    // Go over all permutations of R 
    for (uint64_t R = 1; R <= maxR; R++) {
        // Calculate new Gray code and difference with previous with xor
        uint64_t GrayR = R^(R >> 1);
        uint64_t GrayFlipR = GrayR^prevGrayR;
        prevGrayR = GrayR;
        // Find corresponding flipped index by counting trailing zeros
        int j = __builtin_ctzll(GrayFlipR);
        
        // Change x and y accordingly
        mxComplexDouble productS0 = {1.0, 0};   // Partial sum for S = 0
        mxComplexDouble *xptr = x;              // Pointers for reduced indexing
        mxComplexDouble *yptr = y;              // Pointers for reduced indexing
        mxComplexDouble *xMatptr = &xMat(0, j); // Pointers for reduced indexing        
        if (GrayFlipR&GrayR) {
            // Bit turned on, so add to x
            for (int i = 0; i < N; i++) {
                // x[i] += xMat(i, j);
                *xptr = CompAdd(*xptr, *xMatptr);   
                xMatptr++;
                // productS0 *= x[i];
                productS0 = CompMul(productS0, *xptr);
                xptr++;
            }
            // And y
            for (int k = 0; k < N; k++) {
                for (int i = 0; i < N; i++) {
                    // y(i, k) += W(i, j, k);
                    *yptr = CompAdd(*yptr, W(i, j, k));  
                    yptr++;                     
                }                
            }
        } else {            
            // Bit turned off, so subtract from x
            for (int i = 0; i < N; i++) {
                // x[i] -= xMat(i, j);
                *xptr = CompSub(*xptr, *xMatptr);   
                xMatptr++;
                // productS0 *= x[i];
                productS0 = CompMul(productS0, *xptr);
                xptr++;
            }
            // And y
            for (int k = 0; k < N; k++) {
                for (int i = 0; i < N; i++) {
                    // y(i, k) -= W(i, j, k);
                    *yptr = CompSub(*yptr, W(i, j, k));  
                    yptr++;                     
                }               
            }
        } 
        
        // Go over all permutations of S
        memcpy(xS, x, sizeof(*x)*(size_t)N);    // Allow S to modify x
        prevGrayS = 0;                      
        sumS = productS0;                       // Start with S = {} contribution        
        for (uint64_t S = 1; S <= maxS; S++) {
            // Calculate new Gray code and difference with previous with xor
            uint64_t GrayS = S^(S >> 1);
            uint64_t GrayFlipS = GrayS^prevGrayS;
            prevGrayS = GrayS;
            // Find corresponding flipped index by counting trailing zeros
            int k = __builtin_ctzll(GrayFlipS);

            // Add or subtract flipped row and multiply
            mxComplexDouble product = {1.0, 0};
            mxComplexDouble *xSptr = xS;       // Pointers for reduced indexing
            mxComplexDouble *yptr = &y(0, k);  // Pointers for reduced indexing
            if (GrayFlipS&GrayS) {
                // Bit turned on, so add
                for (int i = 0; i < N; i++) {
                    // x[i] += y(i, k);
                    *xSptr = CompAdd(*xSptr, *yptr); 
                    yptr++;
                    // product *= x[i];
                    product = CompMul(product, *xSptr);
                    xSptr++; 
                }            
            } else {
                // Bit turned off, so subtract
                for (int i = 0; i < N; i++) {
                    // x[i] -= y(i, k);
                    *xSptr = CompSub(*xSptr, *yptr); 
                    yptr++;
                    // product *= x[i];
                    product = CompMul(product, *xSptr);
                    xSptr++; 
                }   
            }
            // Add or subtract product, based on even/odd of S
            if (S&1) {
                sumS = CompSub(sumS, product);
            } else {
                sumS = CompAdd(sumS, product);
            }
        }
        
        // Add or subtract partial sum, based on even/odd of R
        if (R&1) {
            sum = CompSub(sum, sumS);
        } else {
            sum = CompAdd(sum, sumS);
        }        
    }
    
    free(x);
    free(xS);
    free(y);
    free(xMat);
    
    // Final -2 multiplication
    return CompSca(-2.0, sum);
}

/* The gateway routine. */
void mexFunction(int nlhs, mxArray *plhs[], 
                 int nrhs, mxArray const *prhs[]) {

    // Validate number of inputs/outputs
    if (nrhs != 1) {
        mexErrMsgIdAndTxt("perm:rhs",
            "This function requires 1 input argument.");
    } else if (nlhs > 1) {
        mexErrMsgIdAndTxt("perm:lhs",
            "This function only gives 1 output argument.");
    }
    // Validate tensor type
    if (!mxIsDouble(INPUT) || mxIsSparse(INPUT)) {
        mexErrMsgIdAndTxt("perm:inputMatrixType",
            "Input argument must be a full tensor of (complex) doubles.");
    }
    // Validate tensor size
    int DimNum = (int)mxGetNumberOfDimensions(INPUT);
    const mwSize *DimSize = mxGetDimensions(INPUT); 
    int N = (int)DimSize[0];
    if (((DimNum == 2) && (DimSize[0] != DimSize[1]))   // nxm...
     || ((DimNum == 3) && ((DimSize[0] != DimSize[1])   // Not cubic
                        || (DimSize[0] != DimSize[2])
                        || (DimSize[1] != DimSize[2])))
     || (DimNum > 3)){                                  // Too many dimensions        
        mexErrMsgIdAndTxt("perm:inputMatrixSize",
            "Input argument must be a cubic tensor not bigger than 63x63x63.");
    }
        
    #if MX_HAS_INTERLEAVED_COMPLEX
    // Calculate permanent with direct interleaved complex API (>= R2018a)
    if (mxIsComplex(INPUT)) {
        // Initialize complex output and run complex perm function
        OUTPUT = mxCreateDoubleMatrix(1, 1, mxCOMPLEX);        
        if (N == 0) {
            mxGetComplexDoubles(OUTPUT)[0] = (mxComplexDouble) {1.0, 0.0};  // 0x0 tensor has permanent of 1
        } else {
            mxGetComplexDoubles(OUTPUT)[0] = ComplexPermT(mxGetComplexDoubles(INPUT), N);
        }
    } else {
        // Initialize real output and run real perm function
        OUTPUT = mxCreateDoubleMatrix(1, 1, mxREAL);
        if (N == 0) {
            mxGetDoubles(OUTPUT)[0] = 1.0;  // 0x0 matrix has permanent of 1
        } else {            
            mxGetDoubles(OUTPUT)[0] = RealPermT(mxGetDoubles(INPUT), N);
        }
    }
    #else
    // Calculate permanent by first converting to interleaved complex (< R2018a)    
    if (mxIsComplex(INPUT)) {
        // Initialize complex output and run complex perm function
        OUTPUT = mxCreateDoubleMatrix(1, 1, mxCOMPLEX);
        if (N == 0) {
            mxGetPr(OUTPUT)[0] = 1.0;  // 0x0 matrix has permanent of 1
            mxGetPi(OUTPUT)[0] = 0.0;
        } else {
            // Convert matrix to interleaved (negligible overhead)
            mxDouble *WPr = mxGetPr(INPUT);
            mxDouble *WPi = mxGetPi(INPUT);
            mxComplexDouble *W = malloc(sizeof(*W) * (size_t)N * (size_t)N * (size_t)N);
            for (int i = 0; i < N * N * N; i++) {
                W[i].real = WPr[i];
                W[i].imag = WPi[i];
            }
            mxComplexDouble result = ComplexPermT(W, N);
            free(W);
            mxGetPr(OUTPUT)[0] = result.real;
            mxGetPi(OUTPUT)[0] = result.imag;            
        }
    } else {
        // Initialize real output and run real perm function
        OUTPUT = mxCreateDoubleMatrix(1, 1, mxREAL);
        if (N == 0) {            
            mxGetPr(OUTPUT)[0] = 1.0;  // 0x0 matrix has permanent of 1
        } else {
            mxGetPr(OUTPUT)[0] = RealPermT(mxGetPr(INPUT), N);
        }
    }
    #endif
}